package macondo.mak;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import jada.*;
import jada.net.*;

import macondo.util.*;

/**
 *
 *	A very simple prototype of agent.
 *
 *	@author		Andrea Giovannini
 *	@version	beta
 */

public class Agent implements Serializable {

	private String	name;

	private Thread	thread;

	/** Agent properties. */
	private Hashtable prop;

	/** The name of the source place of the agent. */
	private PlaceName source;
	
	/** The name of the current place of the agent. */
	private transient Place currentPlace;
	
	private String	callback;

	/**
	 *	Indicates that the agent is executing for the first time.
	 */
	public static final int		INIT = 0;
		
	/**
	 *	Indicates that this agent is executing after a migration.
	 */
	public static final int		RESTART = 1;
		
	/**
	 *	Indicates that this agent is executing after being made
	 *	persistent.
	 */
	public static final int		RESTORE = 2;

	
   // Constructors ------------------------------------------------------
	
	/** Creates a new <code>Agent</code>. */
	public Agent() {}
	
	/**
	 *	Creates a new <code>Agent</code> with given name.
	 *
	 *	@param	name	the name of the <code>Agent</code>
	 */
	public Agent(String name) {
		this.name = name;
	}
	
	/**
	 *	Creates a new <code>Agent</code> with given name
	 *	and current place name.
	 *
	 *	@param	name	the name of the <code>Agent</code>
	 *	@param	source	the agent's source place name
	 */
	public Agent(String name, PlaceName source) {
		this.name = name;
		this.source = source;
	}
	
	/**
	 *	Creates a new <code>Agent</code> from another one.
	 *
	 *	@param	anAgent		an <code>Agent</code>
	 */
	public Agent(Agent anAgent) {
		this.name = anAgent.name;
		this.source = anAgent.source;
		this.currentPlace = anAgent.currentPlace;
		this.prop = (Hashtable)(anAgent.prop.clone());
	}
	
	// Public Methods ----------------------------------------------------

	public void setName(String name) {
		this.name = name;
	}

	public String getName() {
		return name;
	}

	// Agent properties management methods

	public Object getProperty(Object key) {
		return prop.get(key);
	}
	
	public void putProperty(Object key, Object value) {
		prop.put(key,value);
	}
	
	public Object removeProperty(Object key) {
		return prop.remove(key);
	}

	// Persistence related methods
	
	public void flush() throws IOException {
		currentPlace.flushAgent(this, getClass().getName());
	}
	
	public void flush(Object key) throws IOException {
		currentPlace.flushAgent(this, key);
	}
	
	public void save() throws IOException {
		currentPlace.saveAgent(this, getClass().getName());
	}
	
	public void save(Object key) throws IOException {
		currentPlace.saveAgent(this, key);
	}
	
	// Methods to be defined by the agent programmer

	/**
	 *	Method called to initialize the agent.
	 *	This default implementation does nothing.
	 */
	public void init() {}

	/**
	 *	Method called before agent migration, in order to store relevant
	 *	information in agent state or release resources.
	 *	This default implementation does nothing.
	 */
	public void stop() {}
	

	// Protected Methods -------------------------------------------------
	
	/**
	 *	Moves the agent to a new place.
	 *
	 *	@param	host		destination host name
	 *	@param	port		destination port number
	 *	@param	callback	callback method to execute in new location
	 *	@exception	TransportException			never occurs. One of the
	 *		exception is thrown instead.
	 *	@exception	UnknownDestinationException	occurs if the destination
	 *		place is unknown.
	 *	@exception	TransportFailedException	occurs if migration
	 *		fails for any other reason.
	 */
	protected final void go(String host, int port, String callback)
			throws TransportException {

		this.callback = callback;
		stop();
		try {
			currentPlace.move(this, new PlaceName(host,port));
		} catch (TransportException e) {
			// agent can't be moved
			throw e;
		}
	}
	
	/**
	 *	Moves the agent to a new place.
	 *
	 *	@param	p_name		destination place name
	 *	@param	callback	callback method to execute in new location
	 */
	protected final void go(PlaceName p_name, String callback)
			throws TransportException {
		go(p_name.getHost(), p_name.getPort(), callback);
	}
	
	/**
	 *	Returns current place.
	 *
	 *	@see	macondo.mak.Place
	 */
	protected Place getCurrentPlace() {
		return currentPlace;
	}
	
	/**
	 *	Returns source place name.
	 *
	 *	@return	source place name.
	 *	@see	macondo.mak.PlaceName
	 *	@see	macondo.mak.Place
	 */
	protected PlaceName getSource() {
		return source;
	}
	
	// Package Methods ---------------------------------------------------
	
	final void callback() throws Exception {
		executeMethod(callback, null, null);
	}
	
	/**
	 *	Sets current place.
	 *
	 *	@param	place	new place
	 *	@see	macondo.mak.Place
	 */
	final void setPlace(Place place) {
		currentPlace = place;
	}

	/**
	 *	Sets source place name.
	 *
	 *	@param	aPName	new place name
	 *	@see	macondo.mak.PlaceName
	 *	@see	macondo.mak.Place
	 */
	final void setSource(PlaceName aPName) {
		source = aPName;
	}
	
	/**
	 *	Executes a method on the agent.
	 *
	 *	@param	methodName		the name of the method to execute
	 *	@param	paramTypeList	the method argument parameter list
	 *	@param	argList			the method argument list
	 *	@see	java.lang.reflect.Method
	 */
	final void executeMethod(String methodName, Class[] paramTypeList,
			Object[] argList) throws InvocationTargetException,
			IllegalAccessException, NoSuchMethodException {

		if (methodName == null) return;

		try {
			Method method = this.getClass().getMethod(methodName, paramTypeList);
			try {
				method.invoke(this, argList);
			} catch (InvocationTargetException ite) {
				util.exc(ite);
				ite.getTargetException().printStackTrace();
				throw ite;
			} catch (IllegalAccessException iae) {
				util.exc(iae);
				throw iae;
			}
		} catch (NoSuchMethodException nsme) {
			util.exc(nsme);
			throw nsme;
		}
	}
	
}