package macondo.mak;

import java.util.*;
import java.io.*;

/**
 *	<code>StoreManager</code> is a simple persistent store engine. It
 *	allows to store regular Java objects that implement the
 *	<code>java.io.Serializable</code> interface.
 *	An index table is used to speed up access to stored objects.
 *	The table can also be saved and restored in a different session.
 *	Each object is identified by a <code>key</code> object, that can be
 *	the object application specific name but also any other object.
 *
 *	@author		Andrea Giovannini
 *	@version	beta
 */

public class StoreManager {

	private	Hashtable	table;
	private String		path;
	
	private int			n;

	// Constructor -------------------------------------------------------
	
	/**
	 *	Constructs a new <code>StoreManager</code> with a given path
	 *	where to store objects.
	 *
	 *	@param	path	the file system path where to store objects.
	 */
	public StoreManager(String path) {
		table = new Hashtable();
		this.path = ( path.endsWith("/") ? new String(path) : new String(path + "/") );
		n = 0;
	}
	
	// Public Methods ----------------------------------------------------
	
	/**
	 *	Stores <code>obj</code> and add an entry to the index table.
	 *
	 *	@param		key			the key corresponding to the object.
	 *	@param		obj			the object to store.
	 *	@exception	IOException	if an I/O error occurs writing the object.
	 */
	public synchronized void insertObject(Object key, Object obj) throws IOException {
		String	file_name = key.toString() + nextNum() + ".ser",
				previous;
		
		previous = (String)table.put(key, file_name);
		
		if ( previous != null ) {
			File file = new File( path + previous );
			file.delete();
		}
		
		ObjectOutputStream out = new ObjectOutputStream( new FileOutputStream(path + file_name) );
		out.writeObject( obj );
		out.flush();
	}
	
	/**
	 *	Retrieves the object corresponding to the given <code>key</code>.
	 *
	 *	@param		key			the key used to retrieve the object.
	 *	@return					the object corresponding to the key.
	 *	@exception	IOException	if an I/O error occurs reading the object.
	 *	@exception	ClassNotFoundException	if the class of the object read
	 *										is not found.
	 */
	public synchronized Object getObject(Object key) throws IOException,
			ClassNotFoundException {
		String	file_name = (String)table.get(key);
		Object	obj;
		
		ObjectInputStream in = new ObjectInputStream( new FileInputStream(path + file_name) );
		obj = in.readObject();
		return obj;
	}
	
	/**
	 *	Removes a stored object and the corresponding table entry.
	 *
	 *	@param		key			the key used to retrieve the object.
	 *	@exception	IOException	if some I/O error occurs.
	 */
	public synchronized void removeObject(Object key) throws IOException {
		String	file_name = (String)table.get(key);
		File	file = new File(path + file_name);
		
		file.delete();
		table.remove(key);
	}

	/**
	 *	Saves the index table in stable storage, so it is possible
	 *	to retrieve it in a different session.
	 *
	 *	@exception	IOException		if an I/O error occurs.
	 */
	public synchronized void saveTable() throws IOException {
		ObjectOutputStream out = new ObjectOutputStream( new FileOutputStream(path + "smtable.ser") );
		out.writeObject( table );
		out.flush();
	}
	
	/**
	 *	Loads the index table saved in a previous session.
	 *
	 *	@exception	IOException				if an I/O error occurs.
	 *	@exception	ClassNotFoundException	if a class needed to restore
	 *										the table is not found.
	 */
	public synchronized void loadTable() throws IOException, ClassNotFoundException {
		Object obj;
		
		ObjectInputStream in = new ObjectInputStream( new FileInputStream(path + "smtable.ser") );
		obj = in.readObject();
		
		if ( table != null ) {
			table.clear();
			table = null;
		}
		
		table = (Hashtable)obj;
	}
	
	/**
	 *	Returns an enumeration of the keys in the index table.
	 *
	 *	@return		an enumeration of the keys in the index table.
	 */
	public Enumeration keys() {
		return table.keys();
	}
	
	/**
	 *	Removes all saved objects and clears the index table.
	 */
	public synchronized void clear() {
		// remove all files
		Enumeration elements = table.elements();
		String		file_name;
		File		file;
		
		while( elements.hasMoreElements() ) {
			file_name = (String)elements.nextElement();
			file = new File( path + file_name );
			file.delete();
		}
		
		// if index table is saved to disk, remove it
		file = new File( path + "smtable.ser" );
		if ( file.exists() ) {
			file.delete();
		}
		
		// clear table
		table.clear();
	}
	
	// Private Methods ---------------------------------------------------
	
	/*
	 *	Returns an increasing integer number used to give names to
	 *	objects' files.
	 *
	 *	@return		an increasing integer number.
	 */
	private int nextNum() {
		return n++;
	}
}