package macondo.mjada;

import jada.Tuple;
import jada.JadaException;
import jada.ObjectSpace;

import macondo.util.*;

/**
 *	A <code>Result</code> object is returned as a result of a tuple
 *	request. This way tuples requests are completely decoupled from
 *	the corresponding result fetching.
 *
 *	@author		Andrea Giovannini
 *	@version	beta
 */

public class Result extends Thread {
	public static final int IN = 0;
	public static final int RD = 1;
	public static final Object ERROR = new Object();
	
	private Object			result = null;
	private ObjectSpace		space;
	private Tuple			formal;
	private int				op;
	private Thread			thread;

	// Constructor -------------------------------------------------------
	
	/**
	 *	Builds a new <code>Result</code> object for a tuple request.
	 *
	 *	@param	space	the tuple space where the operation is performed.
	 *	@param	formal	the template tuple.
	 *	@param	op		the operation identifier.
	 */
	public Result(ObjectSpace space, Tuple formal, int op) {
		this.space = space;
		this.formal = formal;
		this.op = op;
		this.thread = new Thread(this);
		thread.start();
	}
	
	// Public Methods ----------------------------------------------------
	
	/**
	 *	Tests if the result is available.
	 *
	 *	@return		<code>true</code> if the result is available,
	 *				<code>false</code> otherwise.
	 */
	public final synchronized boolean isAvailable() {
		return ( result != null );
	}
	
	/**
	 *	Tests if any error occurs while performing the request.
	 *
	 *	@return		<code>true</code> if an error occurs while performing
	 *				the request, <code>false</code> otherwise.
	 */
	public final synchronized boolean isError() {
		return ( result == ERROR );
	}

	/**
	 *	Returns the result tuple of this request. If the result is
	 *	not available this method blocks the calling thread.
	 *
	 *	@return		the result tuple of this request.
	 */
	public final synchronized Object getObject() {
		while( result == null ) {
			try {
				wait();
			} catch (InterruptedException ie) {}
		}
		return result;
	}
	
	/**
	 *	Returns the result tuple of this request. If the result is not
	 *	available waits for the specified timeout. If the timeout value
	 *	is 0 returns the current value of the result.
	 *
	 *	@return		the result tuple of this request.
	 */
	public final synchronized Object getObject(long timeout) {
		if (timeout == 0) {
			return result; // is it correct??
		}
	
		long 	time_to_wait = timeout + System.currentTimeMillis(),
				left;
		
		while( result == null &&
				(left = (time_to_wait - System.currentTimeMillis())) > 0 ) {
			try {
				wait(timeout);
			} catch (InterruptedException ie) {}
		}
		
		return result;
	}
	
	/**
	 *	Kills the current tuple request.
	 */
	public final synchronized void kill() {
		thread.stop();
		if (result != null) {
			space.out(result);
		}
	}
	
	/**
	 *	Performs the real tuple request asynchronously.
	 */
	public void run() {
		Tuple tmp = null;
		Object result = null;
		
		try {
			if ( op == IN ) {
				tmp = (Tuple)space.in(formal);
			} else if ( op == RD ) {
				tmp = (Tuple)space.read(formal);
			}
			result = tmp.getItem(1);
		} catch (JadaException je) {
			util.exc(je);
			result = ERROR;
		}
		
		setObject(result);
	}
	
	// Private Methods ---------------------------------------------------
	
	/**
	 *	Sets the result to the specified object.
	 *
	 *	@param	o	the result object.
	 */
	private final synchronized void setObject(Object o) {
		result = o;
		notifyAll();
	}
}