package macondo.mjada;

import jada.*;
import jada.net.*;
import java.util.*;
import java.io.*;

/**
 *	This class represents the coordination medium used by Macondo
 *	clients. It supports multiple nested tuple spaces with Linda-like
 *	operations.
 *
 *	@author		Andrea Giovannini
 *	@version	beta
 */

public class Space {
	
	private SpaceName	space_name;
	private ObjectSpace	space;
	
	// Constructor -------------------------------------------------------

	/**
	 *	Default constructor. Builds a new local <code>Space</code> object.
	 */	
	public Space() {
		space_name = new SpaceName();
		space = new ObjectSpace();
	}
	
	// Public Methods ----------------------------------------------------

	// Tuple operations

	/**
	 *	Performs a request for withdrawing a tuple from this space.
	 *
	 *	@param	formal	the tuple requested.
	 *	@return			a <code>Result</code> object used as a placeholder
	 *					for the result of this request.
	 *	@see	Result
	 */
	public final Result in(Tuple formal) {
		Tuple match = new Tuple(space_name, formal);
		return new Result(space, match, Result.IN);
	}
	
	/**
	 *	Performs a request for reading a tuple from this space.
	 *
	 *	@param	formal	the tuple requested.
	 *	@return			a <code>Result</code> object uset as a placeholder
	 *					for the result of this request.
	 *	@see	Result
	 */
	public final Result read(Tuple formal) {
		Tuple match = new Tuple(space_name, formal);
		return new Result(space, match, Result.RD);
	}

	/**
	 *	Returns every tuple in this space that matches the given template.
	 *	Matching tuples are removed from the space.
	 *
	 *	@param	formal	the template for requested tuples.
	 *	@return			an enumeration of all tuples in this space
	 *					that match the given template.
	 */
	public final Enumeration inAll(Tuple formal) {
		Tuple	match = new Tuple(space_name, formal),
				result_tuple = null;
		Vector	result = new Vector();
		
		while( (result_tuple = (Tuple)space.in_nb(match)) != null ) {
			result.addElement( result_tuple.getItem(1) );
		}
		
		return result.elements();
	}
	
	/**
	 *	Returns every tuple in this space that matches the given template.
	 *
	 *	@param	formal	the template for requested tuples.
	 *	@return			an enumeration of all tuples in this space
	 *					that match the given template.
	 */
	public final Enumeration readAll(Tuple formal) {
		Tuple	match = new Tuple(space_name, formal),
				result_tuple = null;
		Vector	result = new Vector();
		
		while( (result_tuple = (Tuple)space.in_nb(match)) != null ) {
			result.addElement( result_tuple.getItem(1) );
		}
		
		Enumeration		enum = result.elements();
		
		while( enum.hasMoreElements() ) {
			out( (Tuple)enum.nextElement() );
		}
		
		return enum;
	}

	/**
	 *	Inserts a tuple in this space.
	 *
	 *	@param		tuple			the tuple to insert in this space.
	 *	@exception	SpaceException	thrown if an I/O error occurs.
	 */	
	public final void out(Tuple tuple) {
		SpaceName	sn = new SpaceName();
		sn.cloneItem(space_name);
		Tuple	out_tuple = new Tuple(sn, tuple);
				
		try {		
			space.out(out_tuple);
		} catch (JadaException e) {
			throw new SpaceException();
		}
	}

	// name managing methods
	
	/**
	 *	Name managing method.
	 *	Moves to the referenced tuple space, specified by its name, from
	 *	the current one. Since tuple spaces are nested, names are expressed
	 *	like directory names in a UNIX file system. For instance "/space",
	 *	"/chapter/section" are valid names.
	 *	<p>
	 *	If current tuple space has name "/space1" and this method is
	 *	called with argument "space2", then following operations are performed
	 *	on space "/space1/space2".
	 *	<p>
	 *	Name items can be specified as relative paths but also as absolute
	 *	paths. For instance, if current tuple space is "/working_space" and
	 *	<code>join</code> is called with argument "/games_space", the
	 *	resulting space is "/games_space".
	 *	<p>
	 *	It is also possible moving to the encompassing space specifying
	 *	".." as the <code>join</code> argument.
	 *	<p>
	 *	Finally, a name can have multiple items like "space2/space3".
	 *
	 *	@param	name_items	the tuple space name items.
	 *	@see	SpaceName#join
	 *	@see	#leave()
	 *	@see	#leaveAll()
	 */
	public void join(String name_items) {
		space_name.join(name_items);
	}
	
	/**
	 *	Name managing method.
	 *	Moves to the encompassing space in the coordination structure.
	 *	It is equivalent to <code>join("..")</code>.
	 *
	 *	@see	SpaceName#leave()
	 *	@see	#join
	 *	@see	#leaveAll()
	 */
	public void leave() {
		space_name.leave();
	}
	
	/**
	 *	Name managing method.
	 *	Moves to the root space in the coordination structure.
	 *	It is equivalent to <code>join("/")</code>.
	 *
	 *	@see	SpaceName#leaveAll()
	 *	@see	#join
	 *	@see	#leave()
	 */
	public void leaveAll() {
		space_name.leaveAll();
	}

	/**
	 *	Returns the current tuple space name.
	 *
	 *	@return		the current tuple space name.
	 *	@see		SpaceName
	 */	
	public SpaceName getSpaceName() {
		return space_name;
	}

	// DEBUG METHOD
	public void list() {
		space.print();
	}

	// Package Methods ---------------------------------------------------

	/**
	 *	Sets the Jada object space used for this <code>Space</code>.
	 *
	 *	@param	space	the Jada object space for this <code>Space</code>.
	 */
	void setObjectSpace(ObjectSpace space) {
		this.space = space;
	}
	
	/**
	 *	Returns the Jada object space used for this <code>Space</code>.
	 *
	 *	@return	the Jada object space used for this <code>Space</code>.
	 */
	public ObjectSpace getObjectSpace() {
		return space;
	}

	/**
	 *	Inits the name of this <code>Space</code>.
	 */	
	void initSpaceName() {
		space_name.leaveAll();
	}
	
}
